/*****************************************************************************
	MMPTP.C - Linux USB Kernel Module for PTP digital cameras

	(C) 2002 by Michael Minn - michael@michaelminn.com

	printk messages are written to /var/log/messages

	See /usr/src/linux/Documentation/usb for more information
	about USB programming

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

*******************************************************************************/

/* #include <linux/fs.h> */
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/usb.h>
#include "mmptp.h"

/************************* Globals *****************************/

static purb_t mmptp_urb = NULL;
static int mmptp_in_endpoint = 0;
static int mmptp_out_endpoint = 0;
static int mmptp_interrupt_endpoint = 0;
static wait_queue_head_t mmptp_wait_queue;
static struct usb_device *mmptp_device = NULL;

static void mmptp_wait_callback(urb_t *urb)
{
	if (waitqueue_active(&mmptp_wait_queue))
		wake_up_interruptible(&mmptp_wait_queue);
}

/************************* Linux Device File Operations *****************************/

static int mmptp_file_open (struct inode *inode, struct file *file)
{
	/* int status = 0; */

	if (!mmptp_device)
		return -EHOSTDOWN;

	printk(KERN_INFO "MMPTP opened\n");

	/*
	if ((status = usb_clear_halt(mmptp_device, usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint))) != 0)
	{
		printk(KERN_ERR "MMPTP: failure %d clearing stalled send pipe\n", status);
		return status;
	}

	if ((status = usb_clear_halt(mmptp_device, usb_rcvbulkpipe(mmptp_device, mmptp_in_endpoint))) != 0)
	{
		printk(KERN_ERR "MMPTP: failure %d clearing stalled receive pipe\n", status);
		return status;
	}
	*/

	// MOD_INC_USE_COUNT;

	return 0;
}

static int mmptp_file_release (struct inode *inode, struct file *file)
{
	if (!mmptp_device)
		return -EHOSTDOWN;

	printk(KERN_INFO "MMPTP: file closed\n");

	// MOD_DEC_USE_COUNT;

	return 0;
}

static ssize_t mmptp_file_write (struct file *file, const char *user_data, size_t length, loff_t *offset)
{
	int status = 0;

	/* printk(KERN_INFO "MMPTP: write %d bytes\n", length); */

	if ((!mmptp_device) || (!mmptp_urb))
		return -EHOSTDOWN;

	mmptp_urb->status = 0;
	mmptp_urb->dev = mmptp_device;
	mmptp_urb->transfer_flags = 0;
	mmptp_urb->transfer_buffer_length = length;
	mmptp_urb->complete = mmptp_wait_callback;
	mmptp_urb->pipe = usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint);
	copy_from_user(mmptp_urb->transfer_buffer, user_data, length);

	/*
	if ((status = usb_clear_halt(mmptp_device, usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint))) != 0)
	{
		printk(KERN_ERR "MMPTP: failure %d clearing stalled send pipe\n", status);
		return status;
	}
	*/

	if ((status = usb_submit_urb(mmptp_urb)) != 0)
	{
		printk(KERN_ERR "MMPTP: failure %d submitting write urb\n", status);
		return status;
	}

	interruptible_sleep_on_timeout(&mmptp_wait_queue, HZ * 5);

	if (mmptp_urb->status == USB_ST_URB_PENDING)
	{
		printk(KERN_ERR "MMPTP: Timed out waiting to send command %x\n", (int)((short*)mmptp_urb->transfer_buffer)[3]);
		usb_unlink_urb(mmptp_urb);
		return -ETIMEDOUT;
	}

	if (mmptp_urb->status == USB_ST_STALL)
	{
		/*
		if ((status = usb_clear_halt(mmptp_device, usb_rcvbulkpipe(mmptp_device, mmptp_in_endpoint))) != 0)
			printk(KERN_ERR "MMPTP: failure %d clearing stalled receive pipe\n", status);

		if ((status = usb_clear_halt(mmptp_device, usb_sndbulkpipe(mmptp_device, mmptp_out_endpoint))) != 0)
			printk(KERN_ERR "MMPTP: failure %d clearing stalled send pipe\n", status);
		*/

		printk(KERN_INFO "MMPTP: Device indicated error by stalling endpoint\n");
		return -EPIPE;
	}

	printk(KERN_INFO "MMPTP: Write URB status %d\n", mmptp_urb->status);

	return length;
}

static ssize_t mmptp_file_read (struct file *file, char *data, size_t length, loff_t *offset)
{
	int status = 0;
	int current_length = 0;

	/* printk(KERN_INFO "MMPTP: read %d bytes\n", length); */

	if ((!mmptp_device) || (!mmptp_urb))
		return -EHOSTDOWN;

	do
	{
		mmptp_urb->status = 0;
		mmptp_urb->dev = mmptp_device;
		mmptp_urb->transfer_flags = 0;
		mmptp_urb->transfer_buffer_length = 0x8000;
		/* mmptp_urb->transfer_buffer_length = 0x10000; */
		mmptp_urb->complete = mmptp_wait_callback;
		mmptp_urb->pipe = usb_rcvbulkpipe(mmptp_device, mmptp_in_endpoint);

		if ((status = usb_submit_urb(mmptp_urb)) != 0)
		{
			printk(KERN_ERR "MMPTP: failure %d submitting read urb\n", status);
			return status;
		}

		interruptible_sleep_on_timeout(&mmptp_wait_queue, HZ);

		if (mmptp_urb->status == USB_ST_URB_PENDING)
		{
			printk(KERN_ERR "MMPTP: Timed out waiting for response\n");
			usb_unlink_urb(mmptp_urb);
			return -ETIMEDOUT;
		}

		if (mmptp_urb->status == USB_ST_STALL)
		{
			unsigned short in_error = 0;
			unsigned short out_error = 0;

			/* This code does not work - when invalid object or storage is accessed */
			/* device stalls the endpoint.  Ican't figure out how to clear stall errors */

			printk(KERN_INFO "MMPTP: Device indicated error by stalling endpoint\n");

			if ((status = usb_control_msg(mmptp_device, usb_rcvctrlpipe(mmptp_device, 0),
					USB_REQ_GET_STATUS, USB_DIR_IN | USB_RECIP_ENDPOINT, 0,
					mmptp_in_endpoint, &in_error, 2, HZ * 3)) != 2)
				printk(KERN_ERR "MMPTP: failure %d getting in endpoint status\n", status);

			if ((status = usb_control_msg(mmptp_device, usb_rcvctrlpipe(mmptp_device, 0),
					USB_REQ_GET_STATUS, USB_DIR_IN | USB_RECIP_ENDPOINT, 0,
					mmptp_out_endpoint, &out_error, 2, HZ * 3)) != 2)
				printk(KERN_ERR "MMPTP: failure %d getting out endpoint status\n", status);

			printk(KERN_ERR "MMPTP: Endpoint error status: in %d, out %d\n", in_error, out_error);

			if ((status = usb_control_msg(mmptp_device, usb_sndctrlpipe(mmptp_device, 0),
					USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT, 0, mmptp_in_endpoint, NULL, 0, HZ * 3)) != 0)
				printk(KERN_ERR "MMPTP: failure %d clearing in endpoint\n", status);

			if ((status = usb_control_msg(mmptp_device, usb_sndctrlpipe(mmptp_device, 0),
					USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT, 0, mmptp_out_endpoint, NULL, 0, HZ * 3)) != 0)
				printk(KERN_ERR "MMPTP: failure %d clearing out endpoint\n", status);

			return -EPIPE;
		}

		if ((current_length + mmptp_urb->actual_length) > length)
		{
			printk(KERN_ERR "MMPTP: %d byte read() buffer too small for %d byte message\n", length, mmptp_urb->transfer_buffer_length);
			mmptp_urb->transfer_buffer_length = length;
		}

	/* printk(KERN_INFO "MMPTP: %d bytes received\n", mmptp_urb->transfer_buffer_length); */

		copy_to_user(&data[current_length], mmptp_urb->transfer_buffer, mmptp_urb->actual_length);
		current_length += mmptp_urb->actual_length;
	}
	while (mmptp_urb->actual_length >= mmptp_urb->transfer_buffer_length);

	return current_length;
}

static int mmptp_file_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	printk(KERN_INFO "MMPTP: ioctl %x\n", command);

	return -EINVAL;
}

struct file_operations mmptp_file_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	mmptp_file_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	mmptp_file_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	mmptp_file_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	mmptp_file_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	mmptp_file_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int); */
#endif
};


/************************* Debugging *****************************/

static void mmptp_list_descriptors(struct usb_device *device)
{
	int configuration;
	for (configuration = 0; configuration < device->descriptor.bNumConfigurations; ++configuration)
	{
		int interface;
		printk(KERN_INFO "MMPTP: Configuration %d, %d interfaces\n", configuration, device->config[configuration].bNumInterfaces);

		for (interface = 0; interface < device->config[configuration].bNumInterfaces; ++interface)
		{
			int alternate;
			printk(KERN_INFO "MMPTP:   Interface %d, %d alternates\n", interface, device->config[configuration].interface[interface].num_altsetting);
			for (alternate = 0; alternate < device->config[configuration].interface[interface].num_altsetting; ++alternate)
			{
				int endpoint;
				printk(KERN_INFO "MMPTP:     Alternate %d, %d endpoints\n", interface, device->config[configuration].interface[interface].altsetting[alternate].bNumEndpoints);
				for (endpoint = 0; endpoint < device->config[configuration].interface[interface].altsetting[alternate].bNumEndpoints; ++endpoint)
				{
					printk(KERN_INFO "MMPTP:       Endpoint %d: address %x, max packet %d, ",
						endpoint, device->config[configuration].interface[interface].altsetting[alternate].endpoint[endpoint].bEndpointAddress,
						device->config[configuration].interface[interface].altsetting[alternate].endpoint[endpoint].wMaxPacketSize);

					switch (device->config[configuration].interface[interface].altsetting[alternate].endpoint[endpoint].bmAttributes)
					{
						case USB_ENDPOINT_XFER_CONTROL: printk("Control\n"); break;
						case USB_ENDPOINT_XFER_ISOC: printk("Isosynchronous\n"); break;
						case USB_ENDPOINT_XFER_BULK: printk("Bulk\n"); break;
						case USB_ENDPOINT_XFER_INT: printk("Interrupt\n"); break;
					}
				}
			}
		}
	}
}

/************************* USB Module Registration, Entry, Exit *****************************/


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static void *mmptp_probe(struct usb_device *device, unsigned interface, const struct usb_device_id *id)
#else
static void *mmptp_probe(struct usb_device *device, unsigned interface)
#endif
{
	if (device &&
		 device->config &&
		 device->config->interface &&
		 device->config->interface->altsetting &&
		 (device->config->interface->altsetting->bInterfaceSubClass == 0x01) &&
		 (device->config->interface->altsetting->bInterfaceProtocol == 0x01))
	{
		printk(KERN_INFO "MMPTP: PTP Class Imaging Device Connected - Support not guaranteed\n");
		mmptp_list_descriptors(device);

		mmptp_device = device;
		mmptp_urb = usb_alloc_urb(0);
		mmptp_urb->transfer_buffer = (unsigned char *)kmalloc(0x8000, GFP_KERNEL);

		mmptp_in_endpoint = 0x81;
		mmptp_out_endpoint = 0x01;
		mmptp_interrupt_endpoint = 0x82;

		return device;
	}

	else if ((device->descriptor.idVendor == 0x40a) && (device->descriptor.idProduct == 0x500) && (interface == 0))
	{
		printk(KERN_INFO "MMPTP: Kodak DX-3500 Connected\n");
		mmptp_list_descriptors(device);

		mmptp_device = device;
		mmptp_urb = usb_alloc_urb(0);
		mmptp_urb->transfer_buffer = (unsigned char *)kmalloc(0x8000, GFP_KERNEL);

		mmptp_in_endpoint = 0x81;
		mmptp_out_endpoint = 0x01;
		mmptp_interrupt_endpoint = 0x82;

		return device;
	}

	else if ((device->descriptor.idVendor == 0x40a) && (device->descriptor.idProduct == 0x160) && (interface == 0))
	{
		printk(KERN_INFO "MMPTP: Kodak DC-4800 Connected - Support not guaranteed\n");
		mmptp_list_descriptors(device);

		mmptp_device = device;
		mmptp_urb = usb_alloc_urb(0);
		mmptp_urb->transfer_buffer = (unsigned char *)kmalloc(0x8000, GFP_KERNEL);

		mmptp_in_endpoint = 0x81;
		mmptp_out_endpoint = 0x01;
		mmptp_interrupt_endpoint = 0x82;

		return device;
	}

	else if ((device->descriptor.idVendor == 0x40a) && (device->descriptor.idProduct == 0x121) && (interface == 0))
	{
		printk(KERN_INFO "MMPTP: Kodak DC-240 Connected - Support not guaranteed\n");
		mmptp_list_descriptors(device);

		mmptp_device = device;
		mmptp_urb = usb_alloc_urb(0);
		mmptp_urb->transfer_buffer = (unsigned char *)kmalloc(0x8000, GFP_KERNEL);

		mmptp_in_endpoint = 0x81;
		mmptp_out_endpoint = 0x01;
		mmptp_interrupt_endpoint = 0x82;

		return device;
	}

	return NULL;
}

static void mmptp_disconnect(struct usb_device *device, void *driver_context)
{
	if (mmptp_urb)
	{
		usb_unlink_urb(mmptp_urb);
		kfree(mmptp_urb->transfer_buffer);
		usb_free_urb(mmptp_urb);
		mmptp_urb = NULL;
	}

	mmptp_device = NULL;

	printk(KERN_INFO "MMPTP: Kodak DX3500 Disconnected\n");
}

static struct usb_driver mmptp_driver_registration =
{
	name: "mmptp",
	probe: mmptp_probe,
	disconnect: mmptp_disconnect,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	id_table:	NULL,
	driver_list: {NULL, NULL},
	fops: NULL
#endif
};

int init_module(void)
{
	int status = 0;

	printk(KERN_INFO "MMPTP: Linux USB Module for PTP digital cameras v2002.02.11\n");
	printk(KERN_INFO "MMPTP: Major Number %d\n", MMPTP_MAJOR_NUMBER);

	init_waitqueue_head(&mmptp_wait_queue);

	if ((status = register_chrdev(MMPTP_MAJOR_NUMBER, "mmptp", &mmptp_file_operations)) < 0)
	{
		printk(KERN_ERR "MMPTP: register_chrdev() failed code %d\n", status);
		return -1;
	}

	if ((status = usb_register(&mmptp_driver_registration)) < 0)
		printk(KERN_ERR "MMPTP: Driver failed usb_register with code %d\n", status);

	return 0;
}

int cleanup_module(void)
{
	unregister_chrdev(MMPTP_MAJOR_NUMBER, "mmptp");

	usb_deregister(&mmptp_driver_registration);

	printk(KERN_INFO "MMPTP: Module deregistered and unloaded\n");

	return 0;
}
